# File: OptimalAllocation.R (19/12/2017)
# Coded by: Mathieu Ribatet
#
# Reference:
# Ben Salah, H., De Gooijer, J.G., Gaunnoun, A., and Ribatet, M. 
#   "Mean-variance and mean-semivariance portfolio selection: A
#    multivariate nonparametric approach".
# Accepted for publication in: Financial Markets and Portfolio Management 
##########################################################################
 

dyn.load("Smoothing.so")## or .dll under Windows
library(MASS)
require(quadprog)

markowitz <- function(data, targetRate = 0.01, positive.weights = FALSE){
    data <- as.matrix(data)

    n.assets <- ncol(data)
    varcov <- cov(data)
    mu <- colMeans(data)

    if (positive.weights) {
        ## consider only positive weights
        d <- rep(0, n.assets)
        A <- rbind(rep(1, n.assets), mu, diag(1, n.assets))
        b <- c(1, targetRate, rep(0, n.assets))
        meq <- 1
        opt <- try(solve.QP(2 * varcov, d, t(A), b, meq = meq), silent = TRUE)

        if (class(opt) == "try-error") {
            weights <- rep(NA, n.assets)
            risk <- NA
        } else {
            weights <- opt$solution
            risk <- var(data %*% weights)
        }
    } else {
        ivarcov <- solve(varcov)
        alpha <- as.numeric(rep(1, n.assets) %*% ivarcov %*% rep(1, n.assets))
        lambda <- as.numeric(t(mu) %*% ivarcov %*% rep(1, n.assets))
        theta <- as.numeric(t(mu) %*% ivarcov %*% mu)

        weights <- (alpha * targetRate - lambda) / (alpha * theta - lambda^2) * ivarcov %*% mu +
            (theta - lambda * targetRate) / (alpha * theta - lambda^2) * ivarcov %*% rep(1, n.assets)
        weights <- as.numeric(weights)

        risk <- (alpha * targetRate^2 - 2 * lambda * targetRate + theta) / (alpha * theta - lambda^2)
    }

    names(weights) <- colnames(data)

    return(list(weights = weights, risk = risk))
}

varPort <- function(data, weights)
    return(var(data %*% weights))

semivar <- function(data, weights, bench = 0, negative = TRUE){
    rp <- data %*% weights
    n.obs <- nrow(data)

    if (negative == TRUE) {
        colSums(pmin(rp - bench, 0)^2) / n.obs
    } else {
        colSums(pmax(rp - bench, 0)^2) / n.obs
    }
}

dsr <- function(data, bench = 0, targetRate = NA, tol = 10^-10, positive.weights = FALSE,
                mu = colMeans(data)){

    data <- as.matrix(data)
    n.assets <- ncol(data)
    n.obs <- nrow(data)
    one <- rep(1, n.assets)

    iMold <- diag(10, n.assets)
    weights.old <- one / n.assets

    weights <- one / n.assets
    ##weights <- markowitz(data, targetRate, positive.weights)$weights

    iter <- 1
    if (positive.weights == TRUE){
        d <- rep(0, n.assets)
        A <- rbind(rep(1, n.assets), mu, diag(1, n.assets))
        b <- c(1, targetRate, rep(0, n.assets))
        meq <- 1

        repeat {
            portReturns <- as.numeric(data %*% weights)
            idx <- which(portReturns < bench)
            M <- matrix(0, n.assets, n.assets)
            for (i in idx)
                M <- M + (data[i,] - bench) %*% t(data[i,] - bench)

            M <- M / n.obs
            iM <- ginv(M)

            opt <- try(solve.QP(2 * M, d, t(A), b, meq = meq), silent = TRUE)

            if (class(opt) == "try-error") {
                weights.old <- weights
                weights <- rep(NA, n.assets)
                risk.semiVar <- NA
                break
            } else {
                weights.old <- weights
                weights <- opt$solution
            }

            delta <- sum(abs(M %*% iMold - diag(n.assets)))
            delta <- sum(abs(weights - weights.old))
            ##print(delta)

            if (delta < tol){
                ## Compute one last time all required quantities to compute the DSR
                portReturns <- as.numeric(data %*% weights)
                idx <- which(portReturns < bench)
                M <- matrix(0, n.assets, n.assets)
                for (i in idx)
                    M <- M + (data[i,] - bench) %*% t(data[i,] - bench)

                M <- M / n.obs

                opt <- try(solve.QP(2 * M, d, t(A), b, meq = meq), silent = TRUE)

                if (class(opt) == "try-error") {
                    weights.old <- weights
                    weights <- rep(NA, n.assets)
                    risk.semiVar <- NA
                    break
                } else {
                    weights.old <- weights
                    weights <- opt$solution
                    risk.semiVar <- semivar(data, weights, bench = bench)
                }

                ## Jump outside of the repeat loop
                break
            }

            iMold <- iM
            iter <- iter + 1
        }
    } else {
        repeat {
            returns <- as.numeric(data %*% weights)

            idx <- which(returns < bench)
            M <- matrix(0, n.assets, n.assets)
            for (i in idx)
                M <- M + (data[i,] - bench) %*% t(data[i,] - bench)

            M <- M / n.obs
            iM <- ginv(M)

            alpha <- as.numeric(one %*% iM %*% one)
            lambda <- as.numeric(t(mu) %*% iM %*% one)
            theta <- as.numeric(t(mu) %*% iM %*% mu)

            if (is.na(targetRate)){
                weights.old <- weights
                weights <- iM %*% one / as.numeric(one %*% iM %*% one)
            } else {
                weights.old <- weights
                weights <- (alpha * targetRate - lambda) / (alpha * theta - lambda^2) *
                    iM %*% mu + (theta - lambda * targetRate) / (alpha * theta - lambda^2) *
                    iM %*% one
            }

            ## The stopping criterion depends on the way (or not!)
            ## with smooth portfolio returns so...
            delta <- sum(abs(M %*% iMold - diag(n.assets)))
            ##delta <- sum(abs(weights - weights.old))
            names(weights) <- NULL


            if ((delta < tol) || (iter > 50)){
                ## Compute one last time all required quantities to compute the DSR
                returns <- as.numeric(data %*% weights)

                idx <- which(returns < bench)
                M <- matrix(0, n.assets, n.assets)
                for (i in idx)
                    M <- M + (data[i,] - bench) %*% t(data[i,] - bench)

                M <- M / n.obs
                iM <- ginv(M)

                alpha <- as.numeric(one %*% iM %*% one)
                lambda <- as.numeric(t(mu) %*% iM %*% one)
                theta <- as.numeric(t(mu) %*% iM %*% mu)

                ## Jump outside of the repeat loop
                break
            }

            weights.old <- weights
            iMold <- iM
            iter <- iter + 1

        }

        risk.semiVar <- (alpha * targetRate^2 - 2 * lambda * targetRate + theta) / (alpha * theta - lambda^2)
    }

    risk.var <- as.numeric(as.numeric(weights) %*% cov(data) %*% as.numeric(weights))

    return(list(weights = weights, risk.semiVar = risk.semiVar, risk.var = risk.var, M = M))
}


smoothme <- function(data, h = apply(data, 2, bw.SJ), return.weights = FALSE, which = "mean"){

    if (is.null(dim(data)))
        dim(data) <- c(length(data), 1)

    n.asset <- ncol(data)
    n.obs <- nrow(data)

    if (which == "mean"){
        tmp <- .C("meanSmoothing", as.integer(n.obs), as.integer(n.asset), as.double(unlist(data)),
                  as.double(h), est = double(n.asset * n.obs), weights = double(n.obs^2))

        smoothing.weights <- matrix(tmp$weights, n.obs)
        smoothing.weights <- smoothing.weights / rowSums(smoothing.weights)

        est <- tmp$est
        dim(est) <- c(n.obs, n.asset)
        est <- data.frame(est)
        names(est) <- names(data)

        if (return.weights)
            return(list(est = est, weights = smoothing.weights))

        else return(list(est=est))
    }

    if (which == "median"){
        est <- matrix(NA, n.obs, n.asset)

        data <- t(data)
        for (i in 1:n.obs){
            fun <- function(z)
                sum(sqrt(colSums((data - z)^2)) * exp(-0.5 * colSums(((data - data[,i]) / h)^2)))

            init <- data[,i]
            est[i,] <- optim(init, fun)$par
        }

        est <- data.frame(est)
        names(est) <- rownames(data)

        return(list(est=est))
    }
}

smoothme2 <- function(data, data2, h = bw.SJ(data2), which = "mean", return.weights = FALSE){
    ## A multivariate smoother for data using the conditional median
    ## but the smoothing weights are obtained from data2 and *not* data

    if (is.null(dim(data)))
        dim(data) <- c(length(data), 1)

    n.obs <- nrow(data)
    n.asset <- ncol(data)

    if (which == "mean"){
        tmp <- .C("meanSmoothing2", as.integer(n.obs), as.integer(n.asset), as.double(unlist(data)),
                  as.double(unlist(data2)), as.double(h), est = double(n.asset * n.obs), weights = double(n.obs^2))

        smoothing.weights <- matrix(tmp$weights, n.obs)
        smoothing.weights <- smoothing.weights / rowSums(smoothing.weights)

        est <- tmp$est
        dim(est) <- c(n.obs, n.asset)
        est <- data.frame(est)
        names(est) <- names(data)

        if (return.weights)
            return(list(est = est, weights = smoothing.weights))

        else return(list(est = est))
    }

    if (which == "median"){
        est <- matrix(NA, n.obs, n.asset)
        data <- t(data)

        for (i in 1:n.obs){
            fun <- function(z)
                sum(sqrt(colSums((data - z)^2)) * exp(-0.5 * n.asset * sum(((data2 - data2[i]) / h)^2)))

            init <- data[,i]
            est[i,] <- optim(init, fun)$par
        }

        est <- data.frame(est)
        names(est) <- rownames(data)

        return(list(est=est))
    }
}

adaptivedsr <- function(data, bench = 0, targetRate = NA, tol = 10^-10, positive.weights = FALSE,
                        h = bw.SJ(raw.port.returns), which = "mean"){

    data <- as.matrix(data)
    n.assets <- ncol(data)
    n.obs <- nrow(data)
    one <- rep(1, n.assets)

    iMold <- diag(10, n.assets)
    weights.old <- one / n.assets

    weights <- one / n.assets

    iter <- 1
    if (positive.weights == TRUE){

        d <- rep(0, n.assets)
        b <- c(1, targetRate, rep(0, n.assets))
        meq <- 1

        repeat {
            raw.port.returns <- as.numeric(data %*% weights)
            smooth.port.returns <- as.matrix(smoothme(raw.port.returns, h, which = which)$est)
            smooth.data <- as.matrix(smoothme2(data, raw.port.returns, h, which = which)$est)

            mu <- colMeans(smooth.data)
            A <- rbind(rep(1, n.assets), mu, diag(1, n.assets))

            idx <- which(smooth.port.returns < bench)
            M <- matrix(0, n.assets, n.assets)
            for (i in idx)
                M <- M + (smooth.data[i,] - bench) %*% t(smooth.data[i,] - bench)

            M <- M / n.obs
            iM <- ginv(M)

            opt <- try(solve.QP(2 * M, d, t(A), b, meq = meq), silent = TRUE)

            if (class(opt) == "try-error") {
                weights.old <- weights
                weights <- rep(NA, n.assets)
                risk.semiVar <- NA
                break
            } else {
                weights.old <- weights
                weights <- opt$solution
            }

            delta <- sum(abs(M %*% iMold - diag(n.assets)))
            delta <- sum(abs(weights - weights.old))
            ##print(delta)

            if (delta < tol){
                ## Compute one last time all required quantities to compute the DSR
                raw.port.returns <- as.numeric(data %*% weights)
                smooth.port.returns <- as.matrix(smoothme(raw.port.returns, h, which = which)$est)
                smooth.data <- as.matrix(smoothme2(data, raw.port.returns, h, which = which)$est)

                mu <- colMeans(smooth.data)
                A <- rbind(rep(1, n.assets), mu, diag(1, n.assets))

                idx <- which(smooth.port.returns < bench)
                M <- matrix(0, n.assets, n.assets)
                for (i in idx)
                    M <- M + (smooth.data[i,] - bench) %*% t(smooth.data[i,] - bench)

                Mnew <- M <- M / n.obs
                iM <- ginv(M)

                opt <- try(solve.QP(2 * M, d, t(A), b, meq = meq), silent = TRUE)

                if (class(opt) == "try-error") {
                    weights <- rep(NA, n.assets)
                    risk.semiVar <- NA
                    break
                } else {
                    weights <- opt$solution
                    risk.semiVar <- semivar(smooth.data, weights, bench = bench)
                }

                ## Jump outside of the repeat loop
                break
            }

            iMold <- iM
            iter <- iter + 1
        }
    } else {
        repeat {
            raw.port.returns <- as.numeric(data %*% weights)
            smooth.port.return <- as.matrix(smoothme(raw.port.returns, h, which = which)$est)
            smooth.data <- as.matrix(smoothme2(data, raw.port.returns, h, which = which)$est)
            mu <- colMeans(smooth.data)

            idx <- which(smooth.port.return < bench)
            M <- matrix(0, n.assets, n.assets)
            for (i in idx)
                M <- M + (smooth.data[i,] - bench) %*% t(smooth.data[i,] - bench)

            M <- M / n.obs
            iM <- ginv(M)

            alpha <- as.numeric(one %*% iM %*% one)
            lambda <- as.numeric(t(mu) %*% iM %*% one)
            theta <- as.numeric(t(mu) %*% iM %*% mu)

            if (is.na(targetRate)){
                weights.old <- weights
                weights <- iM %*% one / as.numeric(one %*% iM %*% one)
            } else {
                weights.old <- weights
                weights <- (alpha * targetRate - lambda) / (alpha * theta - lambda^2) *
                    iM %*% mu + (theta - lambda * targetRate) / (alpha * theta - lambda^2) *
                    iM %*% one
            }

            ## The stopping criterion depends on the way (or not!)
            ## with smooth portfolio returns so...
            delta <- sum(abs(M %*% iMold - diag(n.assets)))
            ##delta <- sum(abs(weights - weights.old))
            names(weights) <- NULL


            if ((delta < tol) || (iter > 50)){
                ## Compute one last time all required quantities to compute the DSR
                raw.port.returns <- as.numeric(data %*% weights)
                smooth.port.return <- as.matrix(smoothme(raw.port.returns, h, which = which)$est)
                smooth.data <- as.matrix(smoothme2(data, raw.port.returns, h, which = which)$est)
                mu <- colMeans(smooth.data)

                idx <- which(smooth.port.return < bench)
                M <- matrix(0, n.assets, n.assets)
                for (i in idx)
                    M <- M + (smooth.data[i,] - bench) %*% t(smooth.data[i,] - bench)

                M <- M / n.obs
                iM <- ginv(M)

                alpha <- as.numeric(one %*% iM %*% one)
                lambda <- as.numeric(t(mu) %*% iM %*% one)
                theta <- as.numeric(t(mu) %*% iM %*% mu)

                ## Jump outside of the repeat loop
                break
            }

            weights.old <- weights
            iMold <- iM
            iter <- iter + 1

        }

        risk.semiVar <- (alpha * targetRate^2 - 2 * lambda * targetRate + theta) / (alpha * theta - lambda^2)
    }

    risk.var <- as.numeric(as.numeric(weights) %*% cov(data) %*% as.numeric(weights))

    return(list(weights = weights, risk.semiVar = risk.semiVar, risk.var = risk.var, M = M))
}
