# File: RollingApproach.R (19/12/2017)
# Coded by: Mathieu Ribatet
#
# Reference:
# Ben Salah, H., De Gooijer, J.G., Gaunnoun, A., and Ribatet, M. 
#   "Mean-variance and mean-semivariance portfolio selection: A
#    multivariate nonparametric approach".
# Accepted for publication in: Financial Markets and Portfolio Management 
##########################################################################


rm(list=ls())
library(chron)
library(ggplot2)
library(reshape2)
source("OptimalAllocation.R")
source("utils.R")

data <- read.table("donnees.csv", header = TRUE, sep = ";", dec = ",")
date <- as.Date(data[,1], "%d/%m/%y")

## Remove the date column
data <- data[,-1]

## Keep only the asset which have enough data
asset.removed <- c("Credit.Agricole", "Legrand", "Nokia")
data <- data[,!(names(data) %in% asset.removed)]

## Remove the missing values
## Get the corresponding date
f <- function(x)
    any(is.na(x))

idx.na <- apply(data, 1, f)

data <- data[!idx.na,]
date <- date[!idx.na]

## Revert the data so that it starts with the oldest data
data <- data[nrow(data):1,]
date <- rev(date)


## Work with percentages
data <- 100 * data

## Retrieve the CAC 40
CAC40 <- data[,"CAC40"]
data <- data[,which(names(data) != "CAC40")]

## Do a plot of the CAC40 and highlight calm, good and crisis periods
profit <- function(x)
    cumprod(1 + x / 100)

crisis <- as.Date(c("01/10/07", "31/12/08"), "%d/%m/%y")
calm <- as.Date(c("01/10/03", "31/12/04"), "%d/%m/%y")
good <- as.Date(c("01/10/12", "31/12/13"), "%d/%m/%y")

p <- list()
all.df <- NULL

positive.weights <- FALSE

for (period in c("Calm", "Crisis", "Good")){
    cat("Computations for period", period, "\n")
    if (period == "Crisis")
        idx <- date <= crisis[2] & date >= crisis[1]

    if (period == "Calm")
        idx <- date <= calm[2] & date >= calm[1]

    if (period == "Good")
        idx <- date <= good[2] & date >= good[1]

    sub.data <- data[idx,]
    sub.date <- date[idx]

    df <- data.frame()
    quarters <- quarters(sub.date)
    years <- years(sub.date)
    unique.quarter <- paste("Q", 1:4, sep = "")

    cat("\t Quarter:")
    for (i in c(4,1:3)){##starting from Q4 since then the test period will be Q1
        quarter <- unique.quarter[i]
        cat(quarter, ",")

        train <- which(quarters == quarter & years == max(years))
        data.train <- sub.data[train,]

        if (quarter == "Q4")
            test <- which(quarters == "Q1" & years == max(years))##january of the next year
        else
            test <- which(quarters == unique.quarter[i+1] & years == max(years))

        data.test <- sub.data[test,]
        date.test <- sub.date[test]
        CAC40.test <- CAC40[test]

        ## Smooth the data
        bandwidth <- apply(data.train, 2, bw.SJ)

        univ.mean <- univ.median <- data.train
        for (j in 1:ncol(data.train)){
            univ.mean[,j] <- smoothme(data.train[,j], which = "mean", h = bandwidth[j])$est
            univ.median[,j] <- smoothme(data.train[,j], which = "median", h = bandwidth[j])$est
        }

        multi.mean <- smoothme(data.train, which = "mean", h = bandwidth)$est
        multi.median <- smoothme(data.train, which = "median", h = bandwidth)$est

        ## Computation of the optimal portoflio
        targetRate <- 0.01
        port.naive <- rep(1, ncol(data.train)) / ncol(data.train)
        port.raw <- dsr(data.train, targetRate = targetRate, positive.weights = positive.weights)$weights
        port.univ.mean <- dsr(univ.mean, targetRate = targetRate, positive.weights = positive.weights)$weights
        port.univ.median <- dsr(univ.median, targetRate = targetRate, positive.weights = positive.weights)$weights
        port.multi.mean <- dsr(multi.mean, targetRate = targetRate, positive.weights = positive.weights)$weights
        port.multi.median <- dsr(multi.median, targetRate = targetRate, positive.weights = positive.weights)$weights
        port.adapt.mean <- adaptivedsr(data.train, targetRate = targetRate, positive.weights = positive.weights, which = "mean")$weights
        port.adapt.median <- adaptivedsr(data.train, targetRate = targetRate, positive.weights = positive.weights, which = "median")$weights

        ## Plot the results, i.e., the value of each of the portfolio on the
        ## validation data set
        rate.naive <- profit(as.matrix(data.test) %*% port.naive)
        rate.raw <- profit(as.matrix(data.test) %*% port.raw)
        rate.univ.mean <- profit(as.matrix(data.test) %*% port.univ.mean)
        rate.univ.median <- profit(as.matrix(data.test) %*% port.univ.median)
        rate.multi.mean <- profit(as.matrix(data.test) %*% port.multi.mean)
        rate.multi.median <- profit(as.matrix(data.test) %*% port.multi.median)
        rate.adapt.mean <- profit(as.matrix(data.test) %*% port.adapt.mean)
        rate.adapt.median <- profit(as.matrix(data.test) %*% port.adapt.median)
        rate.CAC40 <- profit(CAC40[test])

        df <- rbind(df,
                    data.frame(Date = date.test, CAC40 = rate.CAC40, Naive = rate.naive,
                               Raw = rate.raw, UMean = rate.univ.mean, MMean = rate.multi.mean,
                               UMedian = rate.univ.median, MMedian = rate.multi.median,
                               AdaptMean = rate.adapt.mean, AdaptMedian = rate.adapt.median))
    }

    cat("\n")

    save(df, file = paste("Results/DataFrameForPeriod", period, "PositiveWeights.RData", sep = ""))

    df <- melt(df, id = "Date", value.name = "Profit", variable.name = "Method")

    all.df <- rbind(all.df, data.frame(df, Period = period))
    tmp <- ggplot(df, aes(x = Date, y = Profit)) + geom_line(aes(colour = Method)) + scale_x_date(date_labels = "%b %d")

    if (period != "Good")
        tmp <- tmp + theme(legend.position = "none")

    ##print(tmp)

    p <- c(p, list(tmp))
}

## We will use a trick to get a nice plot
dummy <- all.df
dummy$Date <- as.Date(format(dummy$Date, "%d / %m"), format = "%d / %m")##problem with February 29th 2008 but we don't care
plotResult <- ggplot(dummy, aes(x = Date, y = Profit, color = Method)) + geom_line() +
    facet_grid(. ~ Period) + scale_x_date(date_labels = "%b %d") +
    geom_hline(yintercept = 1, linetype = 2) +
    theme(legend.position = "top", text = element_text(size = 14), axis.text.x = element_text(angle = 45, hjust = 1))

Sys.setlocale("LC_TIME", "en_US")## To get date in English rather than in French


## Showing the plots
print(plotResult)
